/*
 * Copyright (C) 2010, 2011 Christopher Eby <kreed@kreed.org>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.kreed.vanilla;

import android.app.Activity;
import android.content.Intent;
import android.content.SharedPreferences;
import android.media.AudioManager;
import android.os.Bundle;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.Looper;
import android.os.Message;
import android.os.SystemClock;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.ImageButton;
import android.widget.Toast;

public class PlaybackActivity extends Activity
	implements Handler.Callback,
	           View.OnClickListener,
	           View.OnLongClickListener,
	           CoverView.Callback
{
	public static final int ACTION_NOTHING = 0;
	public static final int ACTION_LIBRARY = 1;
	public static final int ACTION_PLAY_PAUSE = 2;
	public static final int ACTION_NEXT_SONG = 3;
	public static final int ACTION_PREVIOUS_SONG = 4;
	public static final int ACTION_REPEAT = 5;
	public static final int ACTION_SHUFFLE = 6;
	public static final int ACTION_RANDOM = 7;
	public static final int ACTION_ENQUEUE_ALBUM = 8;
	public static final int ACTION_ENQUEUE_ARTIST = 9;
	public static final int ACTION_ENQUEUE_GENRE = 10;
	public static final int ACTION_CLEAR_QUEUE = 11;
	public static final int ACTION_TOGGLE_CONTROLS = 12;

	private int mUpAction;
	private int mDownAction;
	private int mCoverPressAction;
	private int mCoverLongPressAction;

	protected Handler mHandler;
	protected Looper mLooper;

	protected CoverView mCoverView;
	protected ImageButton mPlayPauseButton;

	protected int mState;
	private long mLastStateEvent;
	private long mLastSongEvent;

	@Override
	public void onCreate(Bundle state)
	{
		super.onCreate(state);

		PlaybackService.addActivity(this);

		setVolumeControlStream(AudioManager.STREAM_MUSIC);

		HandlerThread thread = new HandlerThread(getClass().getName());
		thread.start();

		mLooper = thread.getLooper();
		mHandler = new Handler(mLooper, this);
	}

	@Override
	public void onDestroy()
	{
		PlaybackService.removeActivity(this);
		mLooper.quit();
		super.onDestroy();
	}

	@Override
	public void onStart()
	{
		super.onStart();

		if (PlaybackService.hasInstance())
			onServiceReady();
		else
			startService(new Intent(this, PlaybackService.class));

		SharedPreferences prefs = PlaybackService.getSettings(this);
		mUpAction = Integer.parseInt(prefs.getString("swipe_up_action", "0"));
		mDownAction = Integer.parseInt(prefs.getString("swipe_down_action", "0"));
		mCoverPressAction = Integer.parseInt(prefs.getString("cover_press_action", "12"));
		mCoverLongPressAction = Integer.parseInt(prefs.getString("cover_longpress_action", "2"));
	}

	@Override
	public void onResume()
	{
		super.onResume();
		MediaButtonReceiver.registerMediaButton(this);
		MediaButtonReceiver.setInCall(false);
		if (PlaybackService.hasInstance()) {
			PlaybackService service = PlaybackService.get(this);
			service.userActionTriggered();
		}
	}

	@Override
	public boolean onKeyDown(int keyCode, KeyEvent event)
	{
		switch (keyCode) {
		case KeyEvent.KEYCODE_HEADSETHOOK:
		case KeyEvent.KEYCODE_MEDIA_PLAY_PAUSE:
		case KeyEvent.KEYCODE_MEDIA_NEXT:
		case KeyEvent.KEYCODE_MEDIA_PREVIOUS:
			return MediaButtonReceiver.processKey(this, event);
		}

		return super.onKeyDown(keyCode, event);
	}

	@Override
	public boolean onKeyUp(int keyCode, KeyEvent event)
	{
		switch (keyCode) {
		case KeyEvent.KEYCODE_HEADSETHOOK:
		case KeyEvent.KEYCODE_MEDIA_PLAY_PAUSE:
		case KeyEvent.KEYCODE_MEDIA_NEXT:
		case KeyEvent.KEYCODE_MEDIA_PREVIOUS:
			return MediaButtonReceiver.processKey(this, event);
		}

		return super.onKeyUp(keyCode, event);
	}

	public void nextSong()
	{
		setSong(PlaybackService.get(this).nextSong());
	}

	public void previousSong()
	{
		setSong(PlaybackService.get(this).previousSong());
	}

	public void playPause()
	{
		PlaybackService service = PlaybackService.get(this);
		int state = service.playPause();
		if ((state & PlaybackService.FLAG_ERROR) != 0)
			Toast.makeText(this, service.getErrorMessage(), Toast.LENGTH_LONG).show();
		setState(state);
	}

	@Override
	public void onClick(View view)
	{
		switch (view.getId()) {
		case R.id.next:
			nextSong();
			break;
		case R.id.play_pause:
			playPause();
			break;
		case R.id.previous:
			previousSong();
			break;
		case R.id.cover_view:
			performAction(mCoverPressAction);
			break;
		}
	}

	@Override
	public boolean onLongClick(View view)
	{
		if (view.getId() == R.id.cover_view) {
			performAction(mCoverLongPressAction);
			return true;
		}

		return false;
	}

	/**
	 * Called when the PlaybackService state has changed.
	 *
	 * @param state PlaybackService state
	 * @param toggled The flags that have changed from the previous state
	 */
	protected void onStateChange(int state, int toggled)
	{
		if ((toggled & PlaybackService.FLAG_PLAYING) != 0 && mPlayPauseButton != null)
			mPlayPauseButton.setImageResource((state & PlaybackService.FLAG_PLAYING) == 0 ? R.drawable.play : R.drawable.pause);
	}

	protected void setState(final int state)
	{
		mLastStateEvent = SystemClock.uptimeMillis();

		if (mState != state) {
			final int toggled = mState ^ state;
			mState = state;
			runOnUiThread(new Runnable() {
				public void run()
				{
					onStateChange(state, toggled);
				}
			});
		}
	}

	/**
	 * Called by PlaybackService to update the state.
	 */
	public void setState(long uptime, int state)
	{
		if (uptime > mLastStateEvent)
			setState(state);
	}

	/**
	 * Sets up components when the PlaybackService is initialized and available to
	 * interact with. Override to implement further post-initialization behavior.
	 */
	protected void onServiceReady()
	{
		PlaybackService service = PlaybackService.get(this);
		setSong(service.getSong(0));
		setState(service.getState());
	}

	/**
	 * Called when the current song changes.
	 *
	 * @param song The new song
	 */
	protected void onSongChange(Song song)
	{
		if (mCoverView != null)
			mCoverView.querySongs(PlaybackService.get(this));
	}

	protected void setSong(final Song song)
	{
		mLastSongEvent = SystemClock.uptimeMillis();
		runOnUiThread(new Runnable() {
			public void run()
			{
				onSongChange(song);
			}
		});
	}

	/**
	 * Called by PlaybackService to update the current song.
	 */
	public void setSong(long uptime, Song song)
	{
		if (uptime > mLastSongEvent)
			setSong(song);
	}

	/**
	 * Called by PlaybackService to update an active song (next, previous, or
	 * current).
	 */
	public void replaceSong(int delta, Song song)
	{
		if (mCoverView != null)
			mCoverView.setSong(delta + 1, song);
	}

	/**
	 * Called when the content of the media store has changed.
	 */
	public void onMediaChange()
	{
	}

	static final int MENU_PREFS = 2;
	static final int MENU_LIBRARY = 3;
	static final int MENU_SHUFFLE = 4;
	static final int MENU_PLAYBACK = 5;
	static final int MENU_REPEAT = 6;
	static final int MENU_SEARCH = 7;
	static final int MENU_RANDOM = 8;

	@Override
	public boolean onCreateOptionsMenu(Menu menu)
	{
		menu.add(0, MENU_PREFS, 0, R.string.settings).setIcon(R.drawable.ic_menu_preferences);
		menu.add(0, MENU_SHUFFLE, 0, R.string.shuffle_enable).setIcon(R.drawable.ic_menu_shuffle);
		menu.add(0, MENU_REPEAT, 0, R.string.repeat_enable).setIcon(R.drawable.ic_menu_refresh);
		menu.add(0, MENU_RANDOM, 0, R.string.random_enable).setIcon(R.drawable.ic_menu_shuffle);
		return true;
	}

	@Override
	public boolean onPrepareOptionsMenu(Menu menu)
	{
		int state = mState;

		int shuffleRes;
		switch (PlaybackService.shuffleMode(state)) {
		default:
		case SongTimeline.SHUFFLE_NONE:
			shuffleRes = R.string.shuffle_enable;
			break;
		case SongTimeline.SHUFFLE_SONGS:
			shuffleRes = R.string.shuffle_albums;
			break;
		case SongTimeline.SHUFFLE_ALBUMS:
			shuffleRes = R.string.shuffle_disable;
			break;
		}
		menu.findItem(MENU_SHUFFLE).setTitle(shuffleRes);

		int repeatRes;
		if ((state & PlaybackService.FLAG_REPEAT) != 0)
			repeatRes = R.string.repeat_current;
		else if ((state & PlaybackService.FLAG_REPEAT_CURRENT) != 0)
			repeatRes = R.string.repeat_disable;
		else
			repeatRes = R.string.repeat_enable;
		menu.findItem(MENU_REPEAT).setTitle(repeatRes);

		boolean isRandom = (state & PlaybackService.FLAG_RANDOM) != 0;
		// TODO: find icon (dice? arrow pointing in many directions?)
		menu.findItem(MENU_RANDOM).setTitle(isRandom ? R.string.random_disable : R.string.random_enable);

		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item)
	{
		switch (item.getItemId()) {
		case MENU_SHUFFLE:
			cycleShuffle();
			return true;
		case MENU_REPEAT:
			cycleRepeat();
			return true;
		case MENU_RANDOM:
			toggleRandom();
			return true;
		case MENU_PREFS:
			startActivity(new Intent(this, PreferencesActivity.class));
			return true;
		default:
			return false;
		}
	}

	@Override
	public boolean handleMessage(Message msg)
	{
		return false;
	}

	/**
	 * Toggle shuffle mode on/off
	 */
	public void cycleShuffle()
	{
		int state = PlaybackService.get(this).cycleShuffle();
		int shuffleRes;
		switch (PlaybackService.shuffleMode(state)) {
		default:
		case SongTimeline.SHUFFLE_NONE:
			shuffleRes = R.string.shuffle_disabled;
			break;
		case SongTimeline.SHUFFLE_SONGS:
			shuffleRes = R.string.shuffle_songs_enabled;
			break;
		case SongTimeline.SHUFFLE_ALBUMS:
			shuffleRes = R.string.shuffle_albums_enabled;
			break;
		}
		Toast.makeText(this, shuffleRes, Toast.LENGTH_SHORT).show();
		setState(state);
	}

	/**
	 * Cycle repeat mode.
	 */
	public void cycleRepeat()
	{
		int state = PlaybackService.get(this).cycleRepeat();
		int res;
		if ((state & PlaybackService.FLAG_REPEAT) != 0)
			res = R.string.repeat_enabling;
		else if ((state & PlaybackService.FLAG_REPEAT_CURRENT) != 0)
			res = R.string.repeat_current_enabling;
		else
			res = R.string.repeat_disabling;
		Toast.makeText(this, res, Toast.LENGTH_SHORT).show();
		setState(state);
	}

	/**
	 * Toggle random mode on/off
	 */
	public void toggleRandom()
	{
		int state = PlaybackService.get(this).toggleRandom();
		int res = (state & PlaybackService.FLAG_RANDOM) == 0 ? R.string.random_disabling : R.string.random_enabling;
		Toast.makeText(this, res, Toast.LENGTH_SHORT).show();
		setState(state);
	}

	/**
	 * Open the library activity.
	 */
	public void openLibrary()
	{
		startActivity(new Intent(this, LibraryActivity.class));
	}

	public void enqueue(int type)
	{
		PlaybackService.get(this).enqueueFromCurrent(type);
	}

	public void performAction(int action)
	{
		switch (action) {
		case ACTION_NOTHING:
			break;
		case ACTION_LIBRARY:
			openLibrary();
			break;
		case ACTION_PLAY_PAUSE:
			playPause();
			break;
		case ACTION_NEXT_SONG:
			nextSong();
			break;
		case ACTION_PREVIOUS_SONG:
			previousSong();
			break;
		case ACTION_REPEAT:
			cycleRepeat();
			break;
		case ACTION_SHUFFLE:
			cycleShuffle();
			break;
		case ACTION_RANDOM:
			toggleRandom();
			break;
		case ACTION_ENQUEUE_ALBUM:
			enqueue(MediaUtils.TYPE_ALBUM);
			break;
		case ACTION_ENQUEUE_ARTIST:
			enqueue(MediaUtils.TYPE_ARTIST);
			break;
		case ACTION_ENQUEUE_GENRE:
			enqueue(MediaUtils.TYPE_GENRE);
			break;
		case ACTION_CLEAR_QUEUE:
			PlaybackService.get(this).clearQueue();
			Toast.makeText(this, R.string.queue_cleared, Toast.LENGTH_SHORT).show();
			break;
		default:
			throw new IllegalArgumentException("Invalid action: " + action);
		}
	}

	@Override
	public void upSwipe()
	{
		performAction(mUpAction);
	}

	@Override
	public void downSwipe()
	{
		performAction(mDownAction);
	}
}
